from dataclasses import dataclass

from pika.channel import Channel

from myrabbit.core.consumer.listener import Exchange
from myrabbit.core.consumer.listener import Listener
from myrabbit.core.consumer.listener import Queue


@dataclass
class ConsumedChannel:
    listener: Listener
    pika_channel: Channel
    consumer_tag: str = ""

    @property
    def exchange(self) -> Exchange:
        return self.listener.exchange

    @property
    def queue(self) -> Queue:
        return self.listener.queue



from __future__ import annotations

from contextlib import contextmanager
from types import TracebackType
from typing import Any
from typing import Generator
from typing import Optional
from typing import Type

import pika
from pika import BasicProperties
from pika import URLParameters

from myrabbit.core.consumer.pika_message import PikaMessage
from myrabbit.core.publisher.rpc import rpc


class Publisher:
    def __init__(self, connection: pika.BlockingConnection):
        self._connection = connection
        self._channel = self._connection.channel()

    def publish(
        self,
        exchange: str,
        routing_key: str,
        message: bytes,
        properties: Optional[BasicProperties] = None,
    ) -> None:
        self._channel.basic_publish(
            exchange, routing_key, message, properties,
        )

    def close(self) -> None:
        self._channel.close()
        self._connection.close()

    def rpc(
        self,
        exchange: str,
        routing_key: str,
        message: bytes,
        properties: Optional[BasicProperties] = None,
        timeout: Optional[int] = 1,
    ) -> PikaMessage:
        return rpc(
            self._connection, exchange, routing_key, message, properties, timeout
        )

    def __enter__(self) -> Publisher:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_val: Optional[Any],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()


@contextmanager
def make_publisher(amqp_url: str) -> Generator[Publisher, None, None]:
    with pika.BlockingConnection(URLParameters(amqp_url)) as conn:
        with Publisher(conn) as p:
            yield p



import logging
from dataclasses import dataclass

import pika
from pika.channel import Channel
from pika.spec import Basic

from myrabbit.core.consumer.reply import Reply

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class PikaMessage:
    channel: Channel
    basic_deliver: Basic.Deliver
    properties: pika.BasicProperties
    body: bytes

    def requeue(self) -> None:
        logger.info("Requeue message %s", self)
        self.channel.connection.ioloop.add_callback_threadsafe(
            lambda: self.channel.basic_reject(
                self.basic_deliver.delivery_tag, requeue=True
            )
        )

    def acknowledge(self) -> None:
        logger.info("Acknowledging message %s", self)
        self.channel.connection.ioloop.add_callback_threadsafe(
            lambda: self.channel.basic_ack(self.basic_deliver.delivery_tag)
        )

    def reply(self, reply: Reply) -> None:
        if not self.properties.reply_to:
            raise ValueError(
                f"Can not reply to message {self}: "
                f"invalid 'reply_to' value: {self.properties.reply_to!r}"
            )

        reply_rk = self.properties.reply_to
        if reply_rk.startswith("amq."):
            # Direct reply to, exchange must be default.
            reply_exchange = ""
        else:
            # Reply to the exchange message come from.
            reply_exchange = self.basic_deliver.exchange

        logger.info(
            "Replying to exchange %s, routing key %s [%s] with %s",
            reply_exchange,
            reply_rk,
            self.properties.correlation_id,
            reply.body,
        )

        properties = reply.properties or pika.BasicProperties()
        if properties.correlation_id is None:
            properties.correlation_id = self.properties.correlation_id

        self.channel.connection.ioloop.add_callback_threadsafe(
            lambda: self.channel.basic_publish(
                exchange=reply_exchange,
                routing_key=reply_rk,
                body=reply.body,
                properties=properties,
            )
        )



import logging
from dataclasses import dataclass
from typing import Optional

from myrabbit.core.consumer.message_handler import MessageHandler
from . import handle_message_strategy as strategy
from .pika_message import PikaMessage

logger = logging.getLogger(__name__)


@dataclass
class Exchange:
    name: str
    type: str
    durable: bool = True
    auto_delete: bool = True


@dataclass
class Queue:
    name: str
    durable: bool = True
    auto_delete: bool = False
    exclusive: bool = False


@dataclass
class Listener:
    exchange: Exchange
    queue: Queue
    routing_key: str
    handle_message: MessageHandler
    auto_ack: bool = False
    handle_message_strategy: Optional[strategy.HandleMessageStrategy] = None

    def handle(self, message: PikaMessage) -> None:
        execute_strategy = self._get_strategy()
        logger.info("Handling message with %s strategy", execute_strategy)
        execute_strategy(self.handle_message, message)  # type: ignore

    def _get_strategy(self) -> strategy.HandleMessageStrategy:
        if self.handle_message_strategy:
            return self.handle_message_strategy

        return strategy.BaseStrategy(auto_ack=self.auto_ack)



from typing import Callable
from typing import Optional

from myrabbit.core.consumer.reply import Reply
from myrabbit.core.consumer.pika_message import PikaMessage

MessageHandler = Callable[[PikaMessage], Optional[Reply]]



import abc


class ListenEventStrategy(abc.ABC):
    """Strategy that setups event routing.

    The idea and documentation are taken from beautiful Nameko framework.

    https://github.com/nameko/nameko/blob/v3.0.0-rc/nameko/events.py
    """

    @abc.abstractmethod
    def get_queue_name(
        self,
        event_destination: str,
        event_source: str,
        event_name: str,
        method_name: str,
    ) -> str:
        pass



from typing import Any
from typing import Tuple

import orjson

from .serializer import ContentType
from .serializer import Serializer


class JsonSerializer(Serializer):
    def serialize(self, data: Any) -> Tuple[ContentType, bytes]:
        return "application/json", orjson.dumps(data)

    def deserialize(self, data: bytes) -> Any:
        return orjson.loads(data.decode("utf-8"))



from dataclasses import dataclass
from typing import Generic
from typing import TypeVar

from myrabbit.core.consumer.pika_message import PikaMessage

EventType = TypeVar("EventType")


@dataclass
class EventWithMessage(Generic[EventType]):
    event: EventType
    message: PikaMessage



import abc
import logging
from typing import Optional

from myrabbit.core.consumer.message_handler import MessageHandler
from myrabbit.core.consumer.pika_message import PikaMessage
from myrabbit.core.consumer.reply import Reply

logger = logging.getLogger(__name__)


class HandleMessageStrategy(abc.ABC):
    @abc.abstractmethod
    def __call__(self, handle_message: MessageHandler, message: PikaMessage) -> None:
        pass


class BaseStrategy(HandleMessageStrategy):
    def __init__(self, auto_ack: bool):
        self._auto_ack = auto_ack

    def __call__(self, handle_message: MessageHandler, message: PikaMessage) -> None:
        # todo: add retry support on expected exceptions
        try:
            result: Optional[Reply] = handle_message(message)
        except Exception:
            logger.exception("Exception happened during handling message %s", message)
            message.requeue()
            return

        reply_to = message.properties.reply_to

        if not self._auto_ack:
            message.acknowledge()

        if result is not None and reply_to:
            message.reply(result)


class ManualHandle(HandleMessageStrategy):
    def __call__(self, handle_message: MessageHandler, message: PikaMessage) -> None:
        handle_message(message)



from typing import Optional

import pika

from myrabbit.core.publisher import Publisher


class ReconnectingPublisherFactory:
    def __init__(self, amqp_url: str):
        self._amqp_url = amqp_url
        # self._publisher_connection: Optional[pika.BlockingConnection] = None

    # FIXME: use existing connection.
    def get_connection(self) -> pika.BlockingConnection:
        # TODO: implement reconnect on expected exceptions
        #   and reconnection strategy (block, retry times, delays).
        parameters = pika.URLParameters(self._amqp_url)

        # if self._publisher_connection is None or self._publisher_connection.is_closed:
        #     self._publisher_connection = pika.BlockingConnection(parameters)

        # return self._publisher_connection
        return pika.BlockingConnection(parameters)

    def publisher(self) -> Publisher:
        return Publisher(self.get_connection())



from functools import wraps
from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from pika import BasicProperties

from myrabbit.commands.command_bus import CommandBus
from myrabbit.commands.command_outcome import CommandReply
from myrabbit.commands.command_with_message import CommandReplyType
from myrabbit.commands.command_with_message import CommandType
from myrabbit.commands.command_with_message import CommandWithMessage
from myrabbit.commands.command_with_message import ReplyWithMessage
from myrabbit.core.consumer.listener import Listener
from myrabbit.core.converter import Converter
from myrabbit.core.converter import DEFAULT_CONVERTERS


class CommandBusAdapter:
    def __init__(
        self, command_bus: CommandBus, converters: Optional[List[Converter]] = None,
    ):
        self.command_bus = command_bus
        self.converters = converters or DEFAULT_CONVERTERS

    def _converter(self, command: CommandType) -> Converter:
        for converter in self.converters:
            if converter.accepts(command):
                return converter
        raise RuntimeError("Converter not found")

    def send(
        self,
        command_sender: str,
        command_destination: str,
        command: CommandType,
        properties: Optional[BasicProperties] = None,
    ) -> None:
        command_name, body = self._converter(command).name_and_body(command)
        self.command_bus.send(
            command_sender, command_destination, command_name, body, properties
        )

    def listener(
        self,
        command_destination: str,
        command_type: Type[CommandType],
        callback: Callable[[CommandWithMessage], Optional[Union[CommandReply, Any]]],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Listener:
        converter = self._converter(command_type)

        @wraps(callback)
        def instantiate_command(
            command_with_message: CommandWithMessage,
        ) -> Optional[Union[CommandReply, Any]]:
            command_with_message.command = converter.instantiate(
                command_type, command_with_message.command
            )
            return callback(command_with_message)

        command_name = converter.name(command_type)

        return self.command_bus.listener(
            command_destination=command_destination,
            command_name=command_name,
            callback=instantiate_command,
            exchange_params=exchange_params,
            queue_params=queue_params,
        )

    def reply_listener(
        self,
        command_sender: str,
        command_destination: str,
        command_type: Type[CommandType],
        callback: Callable[[ReplyWithMessage], None],
        reply_type: Optional[Type[CommandReplyType]] = None,
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
    ) -> Listener:
        @wraps(callback)
        def instantiate_reply(reply: ReplyWithMessage) -> None:
            if reply_type is not None:
                converter = self._converter(reply_type)
                reply.reply = converter.instantiate(reply_type, reply.reply)
            callback(reply)

        command_name = self._converter(command_type).name(command_type)

        return self.command_bus.reply_listener(
            command_sender=command_sender,
            command_destination=command_destination,
            command_name=command_name,
            callback=instantiate_reply,
            exchange_params=exchange_params,
            queue_params=queue_params,
        )



from typing import List

from myrabbit.core.consumer.consumer import Consumer
from myrabbit.core.consumer.consumer import ThreadedConsumer
from myrabbit.core.consumer.listener import Listener
from myrabbit.core.consumer.reconnecting_consumer import ReconnectingConsumer
from myrabbit.service import Service


def run_services(amqp_url: str, *services: Service) -> None:
    listeners: List[Listener] = sum([s.listeners for s in services], [])
    consumer = ReconnectingConsumer(
        Consumer, consumer_kwargs=dict(amqp_url=amqp_url, listeners=listeners)
    )
    consumer.run()


def run_services_threaded(amqp_url: str, *services: Service) -> None:
    listeners: List[Listener] = sum([s.listeners for s in services], [])
    consumer = ReconnectingConsumer(
        ThreadedConsumer, consumer_kwargs=dict(amqp_url=amqp_url, listeners=listeners)
    )
    consumer.run()



import abc
from typing import Tuple
from typing import Type
from typing import TypeVar

T = TypeVar("T")


class Converter(abc.ABC):
    @abc.abstractmethod
    def name_and_body(self, model: T) -> Tuple[str, dict]:
        pass

    @abc.abstractmethod
    def name(self, model: Type[T]) -> str:
        pass

    @abc.abstractmethod
    def accepts(self, model: T) -> bool:
        pass

    def instantiate(self, model_class: Type[T], body: dict) -> T:
        return model_class(**body)  # type: ignore



from .events import EventBus
from .events import EventWithMessage
from .events import EventBusAdapter
from .commands import CommandBus
from .commands import CommandBusAdapter
from .commands import CommandWithMessage
from .service import Service
from .runner import run_services
from .runner import run_services_threaded



from functools import wraps
from typing import Callable
from typing import List
from typing import Optional
from typing import Type

from pika import BasicProperties

from myrabbit.core.consumer.listener import Listener
from myrabbit.core.converter import Converter
from myrabbit.core.converter import DEFAULT_CONVERTERS
from myrabbit.events.event_bus import EventBus
from myrabbit.events.event_with_message import EventType
from myrabbit.events.event_with_message import EventWithMessage
from myrabbit.events.listen_event_strategy import ListenEventStrategy


class EventBusAdapter:
    def __init__(
        self, event_bus: EventBus, converters: Optional[List[Converter]] = None,
    ):
        self.event_bus = event_bus
        self.converters = converters or DEFAULT_CONVERTERS

    def _converter(self, event: EventType) -> Converter:
        for converter in self.converters:
            if converter.accepts(event):
                return converter
        raise RuntimeError("Converter not found")

    def publish(
        self,
        event_source: str,
        event: EventType,
        properties: Optional[BasicProperties] = None,
    ) -> None:
        event_name, body = self._converter(event).name_and_body(event)
        self.event_bus.publish(event_source, event_name, body, properties)

    def listener(
        self,
        event_destination: str,
        event_source: str,
        event_type: Type[EventType],
        callback: Callable[[EventWithMessage], None],
        exchange_params: Optional[dict] = None,
        queue_params: Optional[dict] = None,
        listen_strategy: Optional[ListenEventStrategy] = None,
        method_name: Optional[str] = None,
    ) -> Listener:
        converter = self._converter(event_type)

        @wraps(callback)
        def instantiate_event(event_with_message: EventWithMessage) -> None:
            event_with_message.event = converter.instantiate(
                event_type, event_with_message.event
            )
            callback(event_with_message)

        event_name = converter.name(event_type)

        return self.event_bus.listener(
            event_destination=event_destination,
            event_source=event_source,
            event_name=event_name,
            callback=instantiate_event,
            exchange_params=exchange_params,
            queue_params=queue_params,
            listen_strategy=listen_strategy,
            method_name=method_name,
        )
